from __future__ import annotations

from collections.abc import Generator
from collections.abc import Iterable
from collections.abc import Mapping
from typing import TYPE_CHECKING
from typing import Any

import pytest
from _pytest._io.saferepr import saferepr
from _pytest.assertion.util import _compare_eq_any

if TYPE_CHECKING:
    from _pytest.config import Config


class UnorderedList(list):
    def __init__(self, expected: Iterable, *, check_type: bool = True) -> None:
        if not isinstance(expected, Iterable):
            msg = f"cannot make unordered comparisons to non-iterable: {expected!r}"
            raise TypeError(
                msg,
            )
        if isinstance(expected, Mapping):
            msg = f"cannot make unordered comparisons to mapping: {expected!r}"
            raise TypeError(msg)
        super().__init__(expected)
        self.expected_type = type(expected) if check_type else None

    def __eq__(self, actual: object) -> bool:
        if self.expected_type is not None and self.expected_type is not type(actual):
            return False
        if not isinstance(actual, Iterable):
            return self.copy() == actual
        actual_list = list(actual)
        if len(actual_list) != len(self):
            return False
        extra_left, extra_right = self.compare_to(actual_list)
        return not extra_left and not extra_right

    def __ne__(self, actual: object) -> bool:
        return not (self == actual)

    def compare_to(self, other: list) -> tuple[list, list]:
        extra_left = list(self)
        extra_right: list[Any] = []
        reordered: list[Any] = []
        placeholder = object()
        for elem in other:
            if elem in extra_left:
                i = extra_left.index(elem)
                reordered.append(extra_left.pop(i))
            else:
                extra_right.append(elem)
                reordered.append(placeholder)
        placeholder_fillers = extra_left.copy()
        for i, elem in reversed(list(enumerate(reordered))):
            if not placeholder_fillers:
                break
            if elem == placeholder:
                reordered[i] = placeholder_fillers.pop()
        self[:] = [e for e in reordered if e is not placeholder]
        return extra_left, extra_right


def unordered(*args: Any, check_type: bool | None = None) -> UnorderedList:
    if len(args) == 1:
        if check_type is None:
            check_type = not isinstance(args[0], Generator)
        return UnorderedList(args[0], check_type=check_type)
    return UnorderedList(args, check_type=False)


def unordered_deep(obj: Any) -> Any:
    if isinstance(obj, dict):
        return {k: unordered_deep(v) for k, v in obj.items()}
    if isinstance(obj, (list, tuple)):
        return unordered(unordered_deep(x) for x in obj)
    return obj


def _compare_eq_unordered(left: Iterable, right: Iterable) -> tuple[list, list]:
    extra_left: list[Any] = []
    extra_right = list(right)
    for elem in left:
        if elem in extra_right:
            extra_right.remove(elem)
        else:
            extra_left.append(elem)
    return extra_left, extra_right


def pytest_assertrepr_compare(
    config: Config,
    op: str,
    left: Any,
    right: Any,
) -> list[str] | None:
    if (isinstance(left, UnorderedList) or isinstance(right, UnorderedList)) and op == "==":
        verbose = config.getoption("verbose")
        left_repr = saferepr(left)
        right_repr = saferepr(right)
        result = [f"{left_repr} {op} {right_repr}"]
        left_type = left.expected_type if isinstance(left, UnorderedList) else type(left)
        right_type = right.expected_type if isinstance(right, UnorderedList) else type(right)
        if left_type and right_type and left_type != right_type:
            result.append("Type mismatch:")
            result.append(f"{left_type} != {right_type}")
        extra_left, extra_right = _compare_eq_unordered(left, right)
        if len(extra_left) == 1 and len(extra_right) == 1:
            result.append("One item replaced:")
            if pytest.version_tuple < (8, 0, 0):  # pragma: no cover
                result.extend(
                    _compare_eq_any(extra_left[0], extra_right[0], verbose=verbose),  # type: ignore[call-arg]
                )
            else:
                result.extend(
                    _compare_eq_any(
                        extra_left[0],
                        extra_right[0],
                        highlighter=config.get_terminal_writer()._highlight,  # noqa: SLF001
                        verbose=verbose,
                    ),
                )
        else:
            if extra_left:
                result.append("Extra items in the left sequence:")
                result.extend(saferepr(item) for item in extra_left)
            if extra_right:
                result.append("Extra items in the right sequence:")
                result.extend(saferepr(item) for item in extra_right)
        return result
    return None



from __future__ import annotations

import collections
from typing import TYPE_CHECKING
from typing import Any
from unittest.mock import ANY

import pytest

from pytest_unordered import UnorderedList
from pytest_unordered import _compare_eq_unordered
from pytest_unordered import unordered
from pytest_unordered import unordered_deep

if TYPE_CHECKING:
    from collections.abc import Iterable
    from collections.abc import Mapping

    from _pytest.pytester import Pytester


@pytest.mark.parametrize(
    ("expected", "actual"),
    [
        (unordered(1, 2, 3), [3, 2, 1]),
        (unordered(1, 2, 3), (3, 2, 1)),
        (unordered(1, 2, 3), {3, 2, 1}),
        (unordered([1, 2, 3]), [3, 2, 1]),
        (unordered((1, 2, 3)), (3, 2, 1)),
        (unordered({1, 2, 3}), {3, 2, 1}),
        (unordered(1, 2, {"a": unordered(4, 5, 6)}), [{"a": [6, 5, 4]}, 2, 1]),
        (unordered([{1: unordered(["a", "b"])}, 2, 3]), [3, 2, {1: ["b", "a"]}]),
        (unordered(x for x in range(3)), [2, 1, 0]),
        (unordered(x for x in range(3)), (2, 1, 0)),
        (unordered(x for x in range(3)), {2, 1, 0}),
        (unordered(x for x in range(3)), range(3)),
        (unordered("abc"), "bac"),
        (unordered("a", "b", "c"), ["b", "a", "c"]),
        (unordered("a", "b", "c"), "bac"),
    ],
)
def test_unordered(expected: UnorderedList, actual: Iterable) -> None:
    assert expected == actual
    assert actual == expected
    assert not (expected != actual)
    assert not (actual != expected)


@pytest.mark.parametrize(
    ("left", "right"),
    [
        (unordered(2, 1, 0), (x for x in range(3))),
        ((x for x in range(3)), unordered(2, 1, 0)),
        (unordered(x for x in range(3)), (2, 1, 0)),
        ((2, 1, 0), unordered(x for x in range(3))),
        (unordered("a", "b", "c"), (x for x in "bac")),
        ((x for x in "bac"), unordered("a", "b", "c")),
    ],
)
def test_unordered_generators(left: Iterable, right: Iterable) -> None:
    # Because general generators can only be consumed once,
    # we can only do one assert
    assert left == right


@pytest.mark.parametrize(
    ("expected", "actual"),
    [
        (unordered([1, 2, 3]), [1, 2, 3, 4]),
        (unordered([1, 2, 3]), [1, 2, 3, 1]),
        (unordered([1, 2, 3]), (1, 2, 3)),
        (unordered([1, 2, 3]), {1, 2, 3}),
        (unordered([1, 2, 3]), (x + 1 for x in range(3))),
        (unordered((1, 2, 3)), (1, 2, 3, 4)),
        (unordered((1, 2, 3)), (1, 2, 3, 1)),
        (unordered((1, 2, 3)), [1, 2, 3]),
        (unordered((1, 2, 3)), {1, 2, 3}),
        (unordered((1, 2, 3)), (x + 1 for x in range(3))),
        (unordered({1, 2, 3}), {1, 2, 3, 4}),
        (unordered({1, 2, 3}), [1, 2, 3]),
        (unordered({1, 2, 3}), (1, 2, 3)),
        (unordered({1, 2, 3}), (x + 1 for x in range(3))),
        (unordered("abc"), ["b", "a", "c"]),
        (unordered("abc"), ("b", "a", "c")),
        (unordered("abc"), {"b", "a", "c"}),
    ],
)
def test_unordered_reject(expected: UnorderedList, actual: Iterable) -> None:
    assert expected != actual
    assert actual != expected
    assert not (expected == actual)
    assert not (actual == expected)


@pytest.mark.parametrize("value", [None, True, 42, object(), type, TypeError])
def test_non_sized_expected(value: Any) -> None:
    with pytest.raises(TypeError, match="cannot make unordered comparisons to non-iterable"):
        UnorderedList(value)


@pytest.mark.parametrize("value", [None, True, 42, object(), type, TypeError])
def test_non_iterable_actual(value: Any) -> None:
    assert not (unordered(1, 2, 3) == value)
    assert not (value == unordered(1, 2, 3))


@pytest.mark.parametrize(
    "value",
    [
        {1: 2, 3: 4},
        collections.defaultdict(int, a=5),
        collections.OrderedDict({1: 2, 3: 4}),
        collections.Counter("count this"),
    ],
)
def test_mapping_expected(value: Mapping) -> None:
    with pytest.raises(TypeError, match="cannot make unordered comparisons to mapping"):
        unordered(value)


@pytest.mark.parametrize("value", [None, type, TypeError])
def test_compare_to_non_sequence(value: Any) -> None:
    assert not unordered("x") == value
    assert unordered("x") != value


def test_check_type() -> None:
    assert not unordered([1]) == {1}
    assert not unordered([1], check_type=True) == {1}
    assert unordered([1], check_type=False) == {1}


@pytest.mark.parametrize(
    ("left", "right", "extra_left", "extra_right"),
    [
        ([1, 2, 3], [1, 2, 3, 4, 5], [], [4, 5]),
        ([3, 2, 1], [1, 2, 3, 4, 5], [], [4, 5]),
        ([3, 2, {1: ["a", "b"]}], [{1: ["a", "b"]}, 2, 3, 4, 5], [], [4, 5]),
        ([3, 2, {1: ["a", "b"]}], [{1: unordered("b", "a")}, 2, 3, 4, 5], [], [4, 5]),
    ],
)
def test_compare_eq_unordered(
    left: Iterable,
    right: Iterable,
    extra_left: list,
    extra_right: list,
) -> None:
    assert _compare_eq_unordered(left, right) == (extra_left, extra_right)


def test_len() -> None:
    assert len(unordered({1: ["a", "b"]}, 2, 3, 4, 5)) == 5


def test_fail_nonunique_left(pytester: Pytester) -> None:
    pytester.makepyfile(
        """
        from pytest_unordered import unordered

        def test_unordered():
            assert unordered(1, 2, 3, 3) == [1, 2, 3]
        """,
    )
    result = pytester.runpytest()
    result.assert_outcomes(failed=1, passed=0)
    result.stdout.fnmatch_lines(
        [
            "E         Extra items in the left sequence:",
            "E         3",
        ],
    )


def test_fail_nonunique_right(pytester: Pytester) -> None:
    pytester.makepyfile(
        """
        from pytest_unordered import unordered

        def test_unordered():
            assert [1, 2, 3] == unordered(1, 2, 3, 3)
        """,
    )
    result = pytester.runpytest()
    result.assert_outcomes(failed=1, passed=0)
    result.stdout.fnmatch_lines(
        [
            "E         Extra items in the right sequence:",
            "E         3",
        ],
    )


def test_replace(pytester: Pytester) -> None:
    pytester.makepyfile(
        """
        from pytest_unordered import unordered

        def test_unordered():
            assert [{"a": 1, "b": 2}, 2, 3] == unordered(2, 3, {"b": 2, "a": 3})
        """,
    )
    result = pytester.runpytest()
    result.assert_outcomes(failed=1, passed=0)
    result.stdout.fnmatch_lines(
        [
            "E         One item replaced:",
            "E         Omitting 1 identical items, use -vv to show",
            "E         Differing items:",
            "E         {'a': 1} != {'a': 3}",
        ],
    )


def test_in(pytester: Pytester) -> None:
    pytester.makepyfile(
        """
        from pytest_unordered import unordered

        def test_unordered():
            assert 1 in unordered(2, 3)
        """,
    )
    result = pytester.runpytest()
    result.assert_outcomes(failed=1, passed=0)
    result.stdout.fnmatch_lines(
        [
            "E       assert 1 in [2, 3]",
            "E        +  where [2, 3] = unordered(2, 3)",
        ],
    )


def test_type_check(pytester: Pytester) -> None:
    pytester.makepyfile(
        """
        from pytest_unordered import unordered

        def test_unordered():
            assert [3, 2, 1] == unordered((1, 2, 3))
        """,
    )
    result = pytester.runpytest()
    result.assert_outcomes(failed=1, passed=0)
    result.stdout.fnmatch_lines(
        [
            "E         Type mismatch:",
            "E         <class 'list'> != <class 'tuple'>",
        ],
    )


def test_reorder_on_eq() -> None:
    unordered_list = unordered([1, 2, 3])
    assert unordered_list == [3, 1, 2]
    assert list(unordered_list) == [3, 1, 2]


def test_mock_any() -> None:
    p_unordered = {"results": unordered({"foo1": ANY}, {"foo2": ANY})}
    test_1 = {"results": [{"foo1": "value10"}, {"foo2": "value20"}]}
    test_2 = {"results": [{"foo1": "value11"}, {"foo2": "value21"}]}
    test_3 = {"results": [{"foo1": "value10"}, {"foo2": "value20"}]}
    assert p_unordered == test_1
    assert p_unordered == test_2
    assert p_unordered == test_3


@pytest.mark.parametrize(
    ("expected", "actual"),
    [
        (unordered_deep([1, 2, 3]), [3, 2, 1]),
        (unordered_deep((1, 2, 3)), (3, 2, 1)),
        (unordered_deep({1, 2, 3}), {3, 2, 1}),
        (unordered_deep([1, 2, {"a": (4, 5, 6)}]), [{"a": [6, 5, 4]}, 2, 1]),  # fmt: skip
        (unordered_deep([{1: (["a", "b"])}, 2, 3]), [3, 2, {1: ["b", "a"]}]),  # fmt: skip
        (unordered_deep(("a", "b", "c")), ["b", "a", "c"]),
    ],
)
def test_unordered_deep(expected: UnorderedList, actual: Iterable) -> None:
    assert expected == actual
    assert actual == expected
    assert not (expected != actual)
    assert not (actual != expected)



#!/usr/bin/env python

from pathlib import Path

from setuptools import setup


def read(fname: str) -> str:
    return (Path(__file__).parent / fname).read_text(encoding="utf-8")


setup(
    name="pytest-unordered",
    version="0.7.0",
    author="Ivan Zaikin",
    author_email="ut@pyngo.tom.ru",
    maintainer="Ivan Zaikin",
    maintainer_email="ut@pyngo.tom.ru",
    license="MIT",
    url="https://github.com/utapyngo/pytest-unordered",
    description="Test equality of unordered collections in pytest",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    packages=["pytest_unordered"],
    package_data={"pytest_unordered": ["py.typed"]},
    install_requires=["pytest>=7.0.0"],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Framework :: Pytest",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Testing",
        "Topic :: Software Development :: Quality Assurance",
        "Topic :: Utilities",
        "Typing :: Typed",
    ],
    entry_points={"pytest11": ["unordered = pytest_unordered"]},
)
